package cache

import (
	"bytes"
	"context"
	"fmt"
	"reflect"
	"time"

	"github.com/dgraph-io/ristretto"
	"github.com/spf13/cast"

	"gitlab.wanzhuangkj.com/tush/xpkg/encoding"
	ctxUtil "gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
	xslice "gitlab.wanzhuangkj.com/tush/xpkg/utils/xslice"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"
)

type memoryCache struct {
	client            *ristretto.Cache
	KeyPrefix         string
	encoding          encoding.Encoding
	DefaultExpireTime time.Duration
	newObject         func() interface{}
}

// NewMemoryCache create a memory cache
func NewMemoryCache(keyPrefix string, encode encoding.Encoding, newObject func() interface{}) Cache {
	// see: https://dgraph.io/blog/post/introducing-ristretto-high-perf-go-cache/
	//		https://www.start.io/blog/we-chose-ristretto-cache-for-go-heres-why/
	config := &ristretto.Config{
		NumCounters: 1e7,     // number of keys to track frequency of (10M).
		MaxCost:     1 << 30, // maximum cost of cache (1GB).
		BufferItems: 64,      // number of keys per Get buffer.
	}
	store, err := ristretto.NewCache(config)
	if err != nil {
		panic("new ristretto cache err:" + err.Error())
	}
	return &memoryCache{
		client:    store,
		KeyPrefix: keyPrefix,
		encoding:  encode,
		newObject: newObject,
	}
}

const logPrefix = "[ristretto]"

// Set data
func (m *memoryCache) Set(ctx context.Context, key string, val interface{}, expiration time.Duration) error {
	buf, err := encoding.Marshal(m.encoding, val)
	if err != nil {
		return xerror.Errorf("encoding.Marshal error: %v, key=%s, val=%+v ", err, key, val)
	}
	if len(buf) == 0 {
		buf = NotFoundPlaceholderBytes
	}
	cacheKey, err := BuildCacheKey(m.KeyPrefix, key)
	if err != nil {
		return xerror.Errorf("BuildCacheKey error: %v, key=%s", err, key)
	}
	st := time.Now()
	ok := m.client.SetWithTTL(cacheKey, buf, 0, expiration)
	logger.Info(fmt.Sprintf("%s set %s", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("val", prettyBytes(buf)), logger.Any("cost", time.Since(st).String()), logger.Any("expiration", expiration.String()), ctxUtil.CtxTraceIDField(ctx))
	if !ok {
		return xerror.New("SetWithTTL failed")
	}

	return nil
}

func (m *memoryCache) IncrBy(ctx context.Context, key string, val int64) (int64, error) {
	cacheKey, err := BuildCacheKey(m.KeyPrefix, key)
	if err != nil {
		return 0, xerror.Errorf("BuildCacheKey error: %v, key=%s", err, key)
	}
	v, ok := m.client.Get(cacheKey)
	if !ok {
		_ = m.client.Set(cacheKey, val, 0)
		return val, nil
	}
	st := time.Now()
	cur := cast.ToInt64(v)
	_ = m.client.Set(cacheKey, cur+val, 0)
	logger.Info(fmt.Sprintf("%s incrby %s", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("cost", time.Since(st).String()), logger.Any("result", val), ctxUtil.CtxTraceIDField(ctx))
	return cur + val, nil
}

// Get data
func (m *memoryCache) Get(ctx context.Context, key string, val interface{}) error {
	cacheKey, err := BuildCacheKey(m.KeyPrefix, key)
	if err != nil {
		return xerror.Errorf("BuildCacheKey error: %v, key=%s", err, key)
	}

	st := time.Now()
	data, ok := m.client.Get(cacheKey)
	if !ok {
		logger.Info(fmt.Sprintf("%s get %s", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("cost", time.Since(st).String()), ctxUtil.CtxTraceIDField(ctx))
		return CacheNotFound
	}

	dataBytes, ok := data.([]byte)
	if !ok {
		return xerror.Errorf("data type error, key=%s, type=%T", key, data)
	}
	if len(dataBytes) == 0 || bytes.Equal(dataBytes, NotFoundPlaceholderBytes) {
		logger.Info(fmt.Sprintf("%s get %s", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("result", "NotFoundPlaceholder"), ctxUtil.CtxTraceIDField(ctx))
		return ErrPlaceholder
	}

	err = encoding.Unmarshal(m.encoding, dataBytes, val)
	if err != nil {
		return xerror.Errorf("encoding.Unmarshal error: %v, key=%s, cacheKey=%s, type=%T, data=%s ",
			err, key, cacheKey, val, dataBytes)
	}
	logger.Info(fmt.Sprintf("%s get %s", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("cost", time.Since(st).String()), logger.Any("result", prettyVal(val)), ctxUtil.CtxTraceIDField(ctx))
	return nil
}

// Del delete data
func (m *memoryCache) Del(ctx context.Context, keys ...string) error {
	if len(keys) == 0 {
		return nil
	}

	key := keys[0]
	cacheKey, err := BuildCacheKey(m.KeyPrefix, key)
	if err != nil {
		return xerror.Errorf("build cache key error, err=%v, key=%s", err, key)
	}
	st := time.Now()
	m.client.Del(cacheKey)
	logger.Info(fmt.Sprintf("%s del %s", logPrefix, cacheKey), logger.Any("cost", time.Since(st).String()), ctxUtil.CtxTraceIDField(ctx))
	return nil
}

// MultiSet multiple set data
func (m *memoryCache) MultiSet(ctx context.Context, valueMap map[string]interface{}, expiration time.Duration) error {
	var err error
	for key, value := range valueMap {
		err = m.Set(ctx, key, value, expiration)
		if err != nil {
			return err
		}
	}
	return nil
}

// MultiGet multiple get data
// deprecated
func (m *memoryCache) MultiGet(ctx context.Context, keys []string, value interface{}) error {
	valueMap := reflect.ValueOf(value)
	var err error
	for _, key := range keys {
		object := m.newObject()
		err = m.Get(ctx, key, object)
		if err != nil {
			continue
		}
		valueMap.SetMapIndex(reflect.ValueOf(key), reflect.ValueOf(object))
	}
	return nil
}

func (m *memoryCache) MGet(ctx context.Context, keys []string) ([]any, error) {
	st := time.Now()
	rs := make([]any, len(keys))
	for i, key := range keys {
		rs[i], _ = m.client.Get(key)
	}
	logger.Info(fmt.Sprintf("%s: mget %s", logPrefix, xslice.Join(keys, " ")), logger.Any("result", rs), logger.Any("cost", time.Since(st).String()), logger.Any("expiration", DefaultNotFoundExpireTime.String()), ctxUtil.CtxTraceIDField(ctx))
	return rs, nil
}

// SetCacheWithNotFound set not found
func (m *memoryCache) SetCacheWithNotFound(ctx context.Context, key string) error {
	cacheKey, err := BuildCacheKey(m.KeyPrefix, key)
	if err != nil {
		return xerror.Errorf("BuildCacheKey error: %v, key=%s", err, key)
	}
	st := time.Now()
	ok := m.client.SetWithTTL(cacheKey, []byte(NotFoundPlaceholder), 0, DefaultNotFoundExpireTime)
	if !ok {
		return xerror.New("SetWithTTL failed")
	}
	logger.Info(fmt.Sprintf("%s: set %s [NotFound]", logPrefix, cacheKey), logger.Any("ok", ok), logger.Any("cost", time.Since(st).String()), logger.Any("expiration", DefaultNotFoundExpireTime.String()), ctxUtil.CtxTraceIDField(ctx))

	return nil
}
